#!/usr/bin/env python3
import os, subprocess, re, time, signal, multiprocessing, socket, sys
from subprocess import CalledProcessError
from args_parser import parse_args
from get_deluge_pass import *


PAUSE_FILE='/dev/shm/PAUSE' # Instead of breaking in the middle of a move, this script can be pause dby creating this file. It moves the current file and pauses.
write_to_file = True # Certain output is written to files for debugging. Set to 'False' if you don't need

#These inputs are expected on the command line but if you want, they can be provided here
deluge_host, deluge_port, deluge_user, FROM, TO = "192.168.1.9", "58884", "myuser", "/home/myuser/Downloads", "/nas/folder" 
#deluge_host, deluge_port, deluge_user = "localhost", "58884", "myuser"

deluge_user = "myuser" # I chose to hardcode this
deluge_pass = get_deluge_pass(deluge_user) # This gets the password from the "~/.config/deluge/auth" file

# Following args are expected on the command line and they override the above ones. Comment out until "deluge_core_command" if you want to keep them hardcoded here
args_n_types = { 'deluge_host': str, 'deluge_port': str, 'FROM': str, 'TO': str}
parsed_args = parse_args(args_n_types)
for arg, value in parsed_args.items():
    #print(arg, value)
    globals()[arg] = value
deluge_core_command = ["/usr/bin/deluge-console", "-d", deluge_host, "-p", deluge_port, "-U", deluge_user, "-P", deluge_pass ]
#print(f'deluge_core: {deluge_core}')
deluge_list_command, ID = [], '' # These are to be populated later
        
def get_tor_list(deluge_list_command: list) -> str:
    try:
        output = subprocess.run(deluge_list_command, check=True, capture_output=True)
        multiline_text_output = output.stdout.decode("utf-8")
        if write_to_file:
            with open('/dev/shm/torrents-list.out', 'w') as f:
                f.writelines(multiline_text_output)
        return multiline_text_output
    except CalledProcessError as cpe:
        print(str(cpe))

def get_size(line): # Get the size of the current torrent
    # Compile the regular expression pattern
    pattern = re.compile(r"Size: (?P<size>\d+(\.\d+)?) (?P<size_unit>[B|G|M]).*")
    # Pattern matching using the "match" method
    if match := pattern.match(line):
        size = float(match.group('size'))
        size_unit = match.group('size_unit')
        return size, size_unit
    else:
        print(f"Invalid line format: '{line}'")

def get_sleeper(size, size_unit): # Rather than querying frequently, sleep some amount of time proportinal to the size of the torrent
    sleeper = 5 # Default value if we fail to calculate
    if size_unit == "G":
        sleeper = size * 4 #Sleeping 4 seconds for each GB worked fine in my setup. Adjust as you wish.
    else:
        sleeper = 3
    return sleeper

def build_torlist_array(multiline_text_output: str) -> list: # This function builds an array from the multiline text response the 'deluge-client' produced
    records = []
    lines = multiline_text_output.splitlines()
    record = {}
    for line in lines:
        line = line.strip()
        if line.startswith(('ID:', 'Download Folder:')):
            key, value = line.split(': ')
            record[key] = value
        elif line.startswith('Size:'):
            record["Size"] = get_size(line)
            record["Sleeper"] = get_sleeper(*record["Size"])
        elif line == '':
            if record:
                records.append({'ID': record.get('ID'), 'Location': record.get('Download Folder'), 'Size': record.get('Size'), 'Sleeper': record.get("Sleeper")})
                record = {}
    if record:
        records.append({'ID': record.get('ID'), 'Location': record.get('Download Folder'), 'Size': record.get('Size'), 'Sleeper': record.get("Sleeper")})

    if write_to_file:
        with open('/dev/shm/torrents-array.out', 'w') as f:
            f.write("\n".join(str(record) for record in records))
    return records

def get_state_n_location(ID: str) -> dict: 
    # Given its ID, returns the sate and the location of a torrent 
    deluge_info_command = deluge_core_command + ["info", ID, "--verbose"]
    state_n_location = {}
    try:
        deluge_info_output = subprocess.run(deluge_info_command, check=True, capture_output=True)
        deluge_info_output_dict = deluge_info_output.stdout.decode("utf-8")
        for line in deluge_info_output_dict.splitlines():
            line = line.strip()
            if line.startswith(('State: Seeding')):
                # Example line with multiple delimiters: "State: Seeding Up Speed: 0.0 K/s"
                state_n_location['State'] = 'Seeding'
            elif line.startswith(('State: Downloading')):
                # Example line with multiple delimiters: "State: Downloading Down Speed: 0.0 K/s Up Speed: 0.0 K/s"
                state_n_location['State'] = 'Downloading'
            elif line.startswith(('State:')):
                if line.count(':') > 1:
                    print(f'Line: "{line}"')
                _, value = line.split(': ')
                state_n_location['State'] = value
            elif line.startswith(('Download Folder:')):
                if line.count(':') > 1:
                    print(f'Line: "{line}"')
                _, value = line.split(': ')
                state_n_location['Location'] = value
        print(str(state_n_location))
        return state_n_location 
    except CalledProcessError as cpe:
        print(str(cpe))

def func_move(deluge_move_process_pid, ID, TO):
    # This moves the torrent
    deluge_move = deluge_core_command + ["move", ID, TO]
    try:
        deluge_move_process = subprocess.Popen(deluge_move)
        deluge_move_process_pid.value = deluge_move_process.pid
    except CalledProcessError as cpe:
        print(str(cpe))
    #print("func_move completed") # This is misleadign because the move has a bug.

def func_mon(deluge_move_process_pid, sleeper, ID, TO):
    # Monitors the progress of func_move and terminates it when necessary
    # deluge-client move operation has a bug. It never completes. Hence it needs to be monitored and killed when the move is complete.
    print(f'Sleeping {sleeper} seconds')
    time.sleep(sleeper) #In seconds. Sleep proportional to the tor size.
    while True:
        current_loc = get_state_n_location(ID)['Location']
        if current_loc == TO:
            # Torrent has been moved
            try:
                if deluge_move_process_pid.value != 0:
                    os.kill(deluge_move_process_pid.value, signal.SIGTERM)
                    print(f"Sent SIGTERM to process {deluge_move_process_pid.value}")
            except OSError:
                break  # Process is no longer running, exit loop
        else:                
            print(f"Current loc: {current_loc}")
            print(f"Moving to  : {TO}")
            time.sleep(10) #In seconds
    print("func_move terminated")

def move_tor(ID, TO, sleeper):
    if get_state_n_location(ID)['Location'] == TO: #For extra safety. Tor may have been moved outside
        print("Source and destination are the same")
    else:
        deluge_move_process_pid = multiprocessing.Value('i', 0) #Shared value
        # Fork and run func_move as a separate process
        p1 = multiprocessing.Process(target=func_move, args=(deluge_move_process_pid, ID, TO))
        p1.start()

        # Run func_mon in the main process
        func_mon(deluge_move_process_pid, sleeper, ID, TO)
        p1.join()

def check_instance(deluge_host, deluge_port):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    result = sock.connect_ex((deluge_host, int(deluge_port)))
    if result == 0:
        print("Instance is running")
        sock.close()
    else:
        print("Instance not running. Exiting...")
        sock.close()
        sys.exit()

if __name__ == "__main__":    # on running the python app
    check_instance(deluge_host, int(deluge_port))
    deluge_list_command = deluge_core_command + ["info", "--verbose" ]
    if deluge_pass is None:
        print("Failed to get the deluge user's password. Exiting")
        sys.exit()
    multiline_text_output = get_tor_list(deluge_list_command)
    torlist_array = build_torlist_array(multiline_text_output)

    for tor in torlist_array:
        if os.path.isfile(PAUSE_FILE):         # Pause if there is the PAUSE file
            print(f'Pause file {PAUSE_FILE} found. Pausing.')
            while os.path.isfile(PAUSE_FILE):  # Wait until the PAUSE file is deleted
                time.sleep(3)
        if tor['Location'] == FROM:
            print(str(tor))
            tor_state = get_state_n_location(tor['ID'])['State']
            if tor_state in ["Checking", "Moving", "Queued", "Error"]:
                print(f"Tor {tor['ID']} is in {tor_state} state. Skipping\n")
            else:
                #input(f'About to move the above tor to {TO}. Press Enter to continue...')
                move_tor(tor['ID'], TO, tor['Sleeper'])
                tor_state = get_state_n_location(tor['ID'])['State']
                if tor_state == "Error":
                    sys.exit()
                while (get_state_n_location(tor['ID'])['State'] in ["Checking", "Moving", "Queued", "Error"]): # Wait until finishes if it went into Checking state or remained in Moving
                    time.sleep(8)
    print("All done")
